unit uCommonTypes;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, ULBLogger;

type

  { THTTPCommandBase }

  THTTPCommandBase = class(TObject)
    protected
      FLibraryName : AnsiString;
      FRequest     : AnsiString;
      FToken       : AnsiString;
    published
      property Lib     : AnsiString read FLibraryName write FLibraryName;
      property Request : AnsiString read FRequest     write FRequest;
      property Token   : AnsiString read FToken       write FToken;
    public
      procedure setRequest(AnObject: TObject);
  end;

  { THTTPRequest }

  THTTPRequest = class(TObject)
  strict private
    FRequest : AnsiString;
    FParams  : AnsiString;

  published
    property Request: String read FRequest write FRequest;
    property Params : String read FParams  write FParams;

  public
    procedure setParams(AnObject: TObject);
  end;

  { THTTPAnswer }

  THTTPAnswer = class(TObject)
    strict private
      FStatus             : Integer;
      FLastError          : UTF8String;
      FData               : UTF8String;
      FStream             : TStream;

      function get_Stream : TStream;

    public
      constructor Create;
      destructor Destroy; override;

      procedure setData(AnObject: TObject; AStream: TStream = nil);

      property Stream             : TStream read get_Stream;

    published
      property Status   : Integer read FStatus write FStatus;
      property LastError: UTF8String read FLastError write FLastError;
      property Data     : UTF8String read FData write FData;
  end;


  IGenericWorker = Interface(IUnknown)
    ['{00C02809-0FF0-E411-830A-543530D8EA6B}']

    function CanManageLogin(): Boolean; stdcall;
    function getToken(): PChar; stdcall;
    function getSessionData(): PChar; stdcall;


    function Execute(RequestString: pChar; SessionData: pChar; pRawData: pByte; RawDataLen: Integer): Boolean; stdcall;
    function getResponseContentType(): pChar; stdcall;
    function getResponseAdditionalHeaders(): pChar; stdcall;

    procedure getAnswerData(out APointer: Pointer; out DataLen: Integer); stdcall;
  end;


  { TGenericWorker }

  TGenericWorker = class(TInterfacedObject, IGenericWorker)
    strict protected
      FLogger : ILBLogger;

      FAnswer : THTTPAnswer;

      FBuffer : array of char;

      FAdditionalHeaders : TStringList;
      FContentType       : String;

      FConfigurationFolder : AnsiString;
      FConfigurationFileName : AnsiString;
      FConfigurationFile: AnsiString;

      function LoadConfiguration(): Boolean; virtual;
      function VerifyData(RequestString: pChar): Boolean;
    public
      constructor Create(); virtual;
      destructor Destroy; override;

      function CanManageLogin(): Boolean; stdcall; virtual;
      function getToken(): PChar; stdcall; virtual;
      function getSessionData(): PChar; stdcall; virtual;


      function Execute(RequestString: pChar; SessionData: pChar; pRawData: pByte; RawDataLen: Integer): Boolean; stdcall; virtual;

      function getResponseContentType(): pChar; stdcall;
      function getResponseAdditionalHeaders(): pChar; stdcall;

      procedure getAnswerData(out APointer: Pointer; out DataLen: Integer); stdcall;

      property ConfigurationFolder : AnsiString write FConfigurationFolder;
      property ConfigurationFileName : AnsiString write FConfigurationFileName;

      property Logger: ILBLogger write FLogger;
  end;

  TGenericWorkerClass = class of TGenericWorker;


const
  HTTP_AnswerStatusOK                 = Integer(200);
  HTTP_AnswerStatusErrorTokenNotFound = Integer(403);
  HTTP_AnswerStatusError              = Integer(404);
  HTTP_AnswerCommandNotExecuted       = Integer(405);

  HTTP_ContentType_JSON = AnsiString('application/json');
  HTTP_ContentType_HTML = AnsiString('text/html');

  HTTP_DummyCommand     = AnsiString('DummyReq'); // It is used for a "keep alive" request


implementation

uses
  ULBUtils;


{ THTTPRequest }

procedure THTTPRequest.setParams(AnObject: TObject);
begin
  if Assigned(AnObject) then
    FParams := AnObject.SerializeAsJSON()
  else
    FParams := '{}';
end;


{ THTTPCommandBase }

procedure THTTPCommandBase.setRequest(AnObject: TObject);
begin
  FRequest := AnObject.SerializeAsJSON();
end;

{ THTTPAnswer }

function THTTPAnswer.get_Stream: TStream;
begin
   Result := FStream;
  if Result <> nil then
    Result.Position := 0;
end;

constructor THTTPAnswer.Create;
begin
  inherited Create;
  FData               := '';
  FStatus             := HTTP_AnswerCommandNotExecuted;
  FStream             := nil;
  FLastError          := '';
end;

destructor THTTPAnswer.Destroy;
begin
  if FStream <> nil then
    FStream.Free;

  inherited Destroy;
end;

procedure THTTPAnswer.setData(AnObject: TObject; AStream: TStream);
begin
  if (AnObject <> nil) then
    FData := AnObject.SerializeAsJSON()
  else
    FData := '';

  FStream := AStream;
  if FStream <> nil then
    FStream.Position := 0;
end;

{ TGenericWorker }

function TGenericWorker.LoadConfiguration: Boolean;
begin
  Result := False;
end;

function TGenericWorker.VerifyData(RequestString: pChar): Boolean;
begin
  Result := False;

  try
    if FLogger <> nil then
      FLogger.Write(5, 'TGenericWorker.VerifyData', lmt_Debug, PChar(Format('Configuring <%s> ...',[Self.ClassName])), GetProcessID);

    if Length(FConfigurationFolder) = 0 then
    begin
      FAnswer.Status := HTTP_AnswerStatusError;
      FAnswer.LastError := 'Configuration folder not set!';

      if FLogger <> nil then
        FLogger.Write(1, 'TGenericWorker.VerifyData', lmt_Warning, PChar(Format('<%s>  -  %s', [Self.ClassName, FAnswer.LastError])), GetProcessID);

      Exit;
    end;

    if not DirectoryExists(FConfigurationFolder) then
    begin
      FAnswer.Status := HTTP_AnswerStatusError;
      FAnswer.LastError := Format('Folder <%s> not found!',[FConfigurationFolder]);

      if FLogger <> nil then
        FLogger.Write(1, 'TGenericWorker.VerifyData', lmt_Warning, PChar(Format('<%s>  -  %s', [Self.ClassName, FAnswer.LastError])), GetProcessID);

      Exit;
    end;

    FConfigurationFolder := IncludeTrailingPathDelimiter(FConfigurationFolder);

    if (RequestString = nil) then
    begin
      FAnswer.Status := HTTP_AnswerStatusError;
      FAnswer.LastError := 'Params not set!';

      if FLogger <> nil then
        FLogger.Write(1, 'TGenericWorker.VerifyData', lmt_Warning, PChar(Format('<%s>  -  %s', [Self.ClassName, FAnswer.LastError])), GetProcessID);

      Exit;
    end;

    FConfigurationFile := FConfigurationFolder + FConfigurationFileName;
    if not FileExists(FConfigurationFile) then
    begin
      FAnswer.Status := HTTP_AnswerStatusError;
      FAnswer.LastError := Format('Configuration file <%s> not found!',[FConfigurationFile]);

      if FLogger <> nil then
        FLogger.Write(1, 'TGenericWorker.VerifyData', lmt_Warning, PChar(Format('<%s>  -  %s', [Self.ClassName, FAnswer.LastError])), GetProcessID);

      Exit;
    end;

    if not Self.LoadConfiguration then
    begin
      FAnswer.Status := HTTP_AnswerStatusError;
      FAnswer.LastError := 'Error reading module configuration';

      if FLogger <> nil then
        FLogger.Write(1, 'TGenericWorker.VerifyData', lmt_Warning, PChar(Format('<%s>  -  %s', [Self.ClassName, FAnswer.LastError])), GetProcessID);

      Exit;
    end;

    Result := True;

  except
    on E: Exception do
    begin
      FAnswer.Status := HTTP_AnswerStatusError;
      FAnswer.LastError := E.Message;

      if FLogger <> nil then
        FLogger.Write(1, 'TGenericWorkerBase.VerifyData', lmt_Error, PChar(Format('<%s>  -  %s', [Self.ClassName, FAnswer.LastError])), GetProcessID);

    end;
  end;
end;

constructor TGenericWorker.Create();
begin
  FConfigurationFile     := '';
  FConfigurationFileName := '';
  FConfigurationFolder   := '';
  FAnswer                := THTTPAnswer.Create;
  FAdditionalHeaders     := TStringList.Create;
  FContentType           := HTTP_ContentType_JSON;
end;

destructor TGenericWorker.Destroy;
begin
  if FLogger <> nil then
    FLogger.Write(3, 'TGenericWorker.Destroy', lmt_Debug, PChar(Format('Destroying <%s> ...',[Self.ClassName])), GetProcessID);

  SetLength(FBuffer, 0);

  FAdditionalHeaders.Free;
  FAnswer.Free;
  FLogger := nil;

  inherited Destroy;
end;

function TGenericWorker.CanManageLogin: Boolean; stdcall;
begin
  Result := False;
end;

function TGenericWorker.getToken: PChar; stdcall;
begin
  Result := nil;
end;

function TGenericWorker.getSessionData: PChar; stdcall;
begin
  Result := nil;
end;

function TGenericWorker.Execute(RequestString: pChar; SessionData: pChar; pRawData: pByte; RawDataLen: Integer): Boolean; stdcall;
begin
  Result := Self.VerifyData(RequestString);

  if not Result then
    FAnswer.Status := HTTP_AnswerStatusError
  else
    FAnswer.Status := HTTP_AnswerStatusOK;
end;

function TGenericWorker.getResponseContentType: pChar; stdcall;
begin
  Result := PChar(FContentType);
end;

function TGenericWorker.getResponseAdditionalHeaders: pChar; stdcall;
begin
  if FLogger <> nil then
    FLogger.Write(7, 'TGenericWorker.getResponseAdditionalHeaders', lmt_Debug, PChar(Format('Additional headers: <%s>', [FAdditionalHeaders.Text])), GetProcessID);

  Result := FAdditionalHeaders.GetText;
end;

procedure TGenericWorker.getAnswerData(out APointer: Pointer; out DataLen: Integer); stdcall;
var
  _sData      : String;
  _JLen       : Integer;
  _StreamSize : Integer = 0;

begin
  try

    _sData := FAnswer.SerializeAsJSON;
    _JLen := Length(_sData);

    FAdditionalHeaders.Values['JSONLEN'] := IntToStr(_JLen);

    if (FAnswer.Stream <> nil) then
      _StreamSize :=  FAnswer.Stream.Size;

    SetLength(FBuffer, _JLen + _StreamSize);
    Move(_sData[1], FBuffer[0], _JLen);

    if _StreamSize > 0 then
    begin
      FAnswer.Stream.Position := 0;
      FAnswer.Stream.ReadBuffer(FBuffer[_JLen], FAnswer.Stream.Size);
    end;

    APointer := @FBuffer[0];
    DataLen := Length(FBuffer);

  except
    on E: Exception do
    begin
      APointer := nil;
      DataLen := 0;

      if FLogger <> nil then
        FLogger.Write(1, 'TGenericWorker.getAnswerData', lmt_Error, PChar(E.Message), GetProcessID);

    end;
  end;
end;

end.

