unit uHTTPCommandExecutor;

{$mode objfpc}{$H+}

interface

// ---------------------------------------------------------------------------------------
// -- Third-party library needed
// 1) Ararat Synapse : http://www.ararat.cz/synapse/doku.php
// ---------------------------------------------------------------------------------------


uses
  Classes, SysUtils, uCommonTypes, ULBLogger;

type
    { THTTPCommandExecutor }

  THTTPCommandExecutor = class(TObject)
    strict private
      FBaseCommand  : THTTPCommandBase;
      FLastError    : AnsiString;
      FAttachment   : TMemoryStream;
      FStreamedData : TMemoryStream;
      FResponseObj  : TObject;
    public
      constructor Create(ACommand: THTTPCommandBase; ResponseObj: TObject);
      function Send(URL: String): Boolean;

      property LastError: AnsiString read FLastError;
      property Attachment: TMemoryStream write FAttachment;
      property StreamedData: TMemoryStream write FStreamedData;
  end;

implementation

uses
  uLBUtils, httpsend, synautil;


var
  HTTP_LastSendTimeStamp : TDateTime = 0;

{ THTTPCommandExecutor }

constructor THTTPCommandExecutor.Create(ACommand: THTTPCommandBase;ResponseObj: TObject);
begin
  inherited Create;
  FAttachment  := nil;
  FResponseObj := ResponseObj;
  FBaseCommand := ACommand;
end;

function THTTPCommandExecutor.Send(URL: String): Boolean;
var
  _HTTP           : THTTPSend = nil;
  _Response       : TStringList = nil;
  _Answer         : THTTPAnswer = nil;

  _JSonString     : AnsiString;

  _Len : Integer;

  _RawDataLen : Integer;

begin
  Result := False;
  FLastError := 'Command not executed!';

  if FBaseCommand = nil then
  begin
    FLastError := 'Base command not set!';
    LBLogger.Write(1, 'THTTPCommandExecutor.Send', lmt_Warning, PChar(FLastError), GetProcessID);
    Exit;
  end;

  if Length(URL) = 0 then
  begin
    FLastError := 'Destination URL not set!';
    LBLogger.Write(1, 'THTTPCommandExecutor.Send', lmt_Warning, PChar(FLastError), GetProcessID);
    Exit;
  end;

  _JsonString := FBaseCommand.SerializeAsJSON();

  if Length(_JsonString) = 0 then
  begin
    LBLogger.Write(1, 'THTTPCommandExecutor.Send', lmt_Warning, 'No data to send!', GetProcessID);
    Exit;
  end;

  try
    _HTTP := THTTPSend.Create;
    WriteStrToStream(_HTTP.Document, _JSonString);

    if FAttachment <> nil then
    begin
      LBLogger.Write(3, 'THTTPCommandExecutor.Send', lmt_Debug, 'Inserting attachment ...', GetProcessID);
      FAttachment.Seek(0, soFromBeginning);
      _HTTP.Document.WriteBuffer(FAttachment.Memory^, FAttachment.Size);
      _HTTP.MimeType := 'Application/octet-stream';
    end
    else
      _HTTP.MimeType := 'application/json; charset=utf-8';

    _HTTP.Protocol := '1.1';
    _HTTP.KeepAlive := True;
    Result := _HTTP.HTTPMethod('POST', URL + '?JLen=' + IntToStr(Length(_JSonString)));

    if Result then
    begin
      HTTP_LastSendTimeStamp := Now();

      _HTTP.Document.Seek(0, soFromBeginning);

      if _HTTP.Document.Size = 0 then
      begin
        Result := False;
        FLastError := 'No answer data';
        LBLogger.Write(1, 'THTTPCommandExecutor.Send', lmt_Warning, PChar(FLastError), GetProcessID);
      end
      else begin
        LBLogger.Write(5, 'THTTPCommandExecutor.Send', lmt_Debug, PChar(Format('Document size: %d',[_HTTP.Document.Size])), GetProcessID);

        _HTTP.Headers.NameValueSeparator := ':';
        _Len := StrToIntDef(Trim(_HTTP.Headers.Values['JSONLEN']), 0);

        _RawDataLen := _HTTP.Document.Size - _Len;
        if _Len > 0 then
        begin
          SetLength(_JSonString, _Len);
          _HTTP.Document.Read(_JSonString[1], _Len);

          _Answer := THTTPAnswer.Create;

          if _Answer.FillFromJSON(_JSonString) then
            Result := Boolean(_Answer.Status = HTTP_AnswerStatusOK)
          else
            LBLogger.Write(1, 'THTTPCommandExecutor.Send', lmt_Warning, PChar(Format('Error decoding received data: %s',[_JSonString])), GetProcessID);


          if Result then
          begin
            LBLogger.Write(7, 'THTTPCommandExecutor.Send', lmt_Debug, PChar('Answer data: ' + _Answer.Data), GetProcessID);

            if (FResponseObj <> nil) and (Length(_Answer.Data) > 0) then
            begin
              LBLogger.Write(5, 'THTTPCommandExecutor.Send', lmt_Debug, PChar(Format('Result type <%s>',[FResponseObj.ClassName])), GetProcessID);


              if not FResponseObj.FillFromJSON(_Answer.Data) then
                LBLogger.Write(1, 'THTTPCommandExecutor.Send', lmt_Warning, PChar(Format('Error filling response object: <%s>', [FResponseObj.ClassName])), GetProcessID);
            end;
          end
          else begin
            LBLogger.Write(7, 'THTTPCommandExecutor.Send', lmt_Warning, PChar(Format('Response error: <%s>', [_JSonString])), GetProcessID);
            FLastError := _Answer.LastError;
          end;
        end
        else
          LBLogger.Write(1, 'THTTPCommandExecutor.Send', lmt_Warning, 'No JSON DATA', GetProcessID);

        if (_RawDataLen > 0) then
        begin
          LBLogger.Write(5, 'THTTPCommandExecutor.Send', lmt_Debug, PChar(Format('Retrieve raw data. Size: %d',[_RawDataLen])), GetProcessID);

          if (FStreamedData <> nil) then
          begin
            FStreamedData.SetSize(_RawDataLen);
            _HTTP.Document.ReadBuffer(FStreamedData.Memory^,_RawDataLen);
          end
          else begin
            Result := False;
            LBLogger.Write(1, 'THTTPCommandExecutor.Send', lmt_Warning, 'Streamed data stream not set!', GetProcessID);
          end;
        end;

        if (_RawDataLen = 0) and (_Len = 0) then
        begin
          Result := False;
          LBLogger.Write(1, 'THTTPCommandExecutor.Send', lmt_Warning, 'No DATA RECEIVED!', GetProcessID);
        end;
      end;
    end
    else begin
      FLastError := 'HTTP error code: ' + IntToStr(_HTTP.ResultCode) + '  -  ' + _HTTP.ResultString;
      LBLogger.Write(1, 'THTTPCommandExecutor.Send', lmt_Warning, PChar('Request: ' + URL + '?JLen=' + IntToStr(Length(_JSonString)) + '  -  Error: ' + FLastError), GetProcessID);
      LBLogger.Write(5, 'THTTPCommandExecutor.Send', lmt_Debug, PChar('Request: ' + _JSonString), GetProcessID);
    end;

  except
    on E: Exception do
    begin
      Result := False;
      FLastError := Format('Error: ',[E.Message]);
      LBLogger.Write(1, 'THTTPCommandExecutor.Send', lmt_Error, PChar(FLastError), GetProcessID);
    end;
  end;

  if _HTTP <> nil then
    FreeAndNil(_HTTP);

  if _Answer <> nil then
    FreeAndNil(_Answer);

  if _Response <> nil then
    FreeAndNil(_Response);

  if Result then
    FLastError := '';
end;


end.

