unit ufcpProxy;

{$mode objfpc}{$H+}


interface

uses
  SysUtils, Classes, httpdefs, fpHTTP, fpWeb, uLibrariesManager, uTokenManager, ULBLogger, uCommonTypes;


// --------------------------------------------------------------------------------------
// ---------- httpd.conf changes

// -- Load the proxy module
// LoadModule mod_fpcproxy <module path for example modules/libmod_fpcproxy.so>

// -- Tell httpd when it must use the module
// <Location /Commands>
//    SetHandler mod_fpcproxy
// </Location>

// --------------------------------------------------------------------------------------

type

  { TFPWebProxy }

  TFPWebProxy = class(TFPWebModule)
    procedure DataModuleCreate(Sender: TObject);
    procedure DataModuleDestroy(Sender: TObject);
    procedure DataModuleRequest(Sender: TObject; ARequest: TRequest; AResponse: TResponse; var Handled: Boolean);

  private
    FLibrariesManager : TLibrariesManager;

    procedure SendAnswer(AResponse: TResponse; AnAnswer: THTTPAnswer);
    procedure getDataSent(ARequest: TRequest; out JSONData: String; out AdditionalData: String);
  public
  end;

var
  FPWebProxy: TFPWebProxy = nil;
  InstanceCounter : Integer = 0;

implementation

{$R *.lfm}

uses
  FileUtil, uLBUtils, UGlobals;


{ TFPWebProxy }

procedure TFPWebProxy.DataModuleCreate(Sender: TObject);
begin
  try
    Inc(InstanceCounter);
    if InstanceCounter = 1 then
    begin
      InitializeFolders();
      InitializeLogger();
      InitializeConfigurationData();
      InitializeTokenManager();
      InitializeLibrariesManager();
      LBLogger.Write(3, 'TFPWebProxy.DataModuleCreate', lmt_Info, '------> Initialization completed! <------', GetProcessID);
    end;

    LBLogger.Write(3, 'TFPWebProxy.DataModuleCreate', lmt_Info, PChar(Format('------> Data module created  -  Instance counter: %d <------',[InstanceCounter])), GetProcessID);

    FLibrariesManager := GlobalLibrariesManager;

  except
    on E: Exception do
      LBLogger.Write(1, 'TFPWebProxy.DataModuleCreate', lmt_Error, PChar(E.Message), GetProcessID);
  end;
end;

procedure TFPWebProxy.DataModuleDestroy(Sender: TObject);
begin
  try
    LBLogger.Write(3, 'TFPWebProxy.DataModuleDestroy', lmt_Info, PChar(Format('------> Destroying data module  -  Instance counter: %d <------',[InstanceCounter])), GetProcessID);

    if InstanceCounter = 1 then
    begin
      LBLogger.Write(1, 'TFPWebProxy.DataModuleDestroy', lmt_Debug, 'Destroying global data ...', GetProcessID);
      DestroyGlobals();
    end;

    Dec(InstanceCounter);
  except

  end;
end;

procedure TFPWebProxy.DataModuleRequest(Sender: TObject; ARequest: TRequest; AResponse: TResponse; var Handled: Boolean);
var
  _Answer            : THTTPAnswer = nil;
  _JSONData          : AnsiString = '';
  _AdditionalData    : AnsiString = '';
  _AdditionalDataLen : Integer;
  _ARequest          : THTTPCommandBase = nil;
  _Executor          : IGenericWorker = nil;
  _SessionData       : AnsiString = '';
  _Token             : AnsiString = '';
  _sError            : AnsiString = '';
  _Data              : Pointer = nil;
  _DataLen           : Integer = 0;
  _pAdditionalData   : pByte = nil;
  _Error             : Boolean = false;

  _Response          : TStream = nil;

  _AddHeaders        : String = '';

begin
  LBLogger.Write(7, 'TFPWebProxy.DataModuleRequest', lmt_Debug, '------> New request <------', GetProcessID);

  Handled := True;

  try

    Self.getDataSent(ARequest, _JSONData, _AdditionalData);

    if (Length(_JSONData) = 0) then // Answer error: no data received!
    begin
      LBLogger.Write(1, 'TFPWebProxy.DataModuleRequest', lmt_Warning, 'No data received!', GetProcessID);
      _Answer := THTTPAnswer.Create;
      _Answer.LastError := 'No data received!';
      Self.SendAnswer(AResponse, _Answer);
      FreeAndNil(_Answer);
      Exit;
    end;

    _ARequest := THTTPCommandBase.Create;
    if not _ARequest.FillFromJSON(_JSONData) then
    begin
      LBLogger.Write(1, 'TFPWebProxy.DataModuleRequest', lmt_Warning, PChar(Format('Error decoding command: <%s>', [_JSONData])), GetProcessID);
      _Answer := THTTPAnswer.Create;
      _Answer.Status := HTTP_AnswerStatusError;
      _Answer.LastError := 'Error decoding request';
      Self.SendAnswer(AResponse, _Answer);
      FreeAndNil(_Answer);
    end
    else begin

      LBLogger.Write(7, 'TFPWebProxy.DataModuleRequest', lmt_Debug, PChar(Format('Request: %s', [_ARequest.Lib])), GetProcessID);

      if _ARequest.Lib = HTTP_DummyCommand then // Answer to a DUMMY REQUEST (keep alive)
      begin
        _Answer := THTTPAnswer.Create;
        _Answer.Status := HTTP_AnswerStatusOK;
        Self.SendAnswer(AResponse, _Answer);
        FreeAndNil(_Answer);
        FreeAndNil(_ARequest);
        Exit;
      end;

      LBLogger.Write(7, 'TFPWebProxy.DataModuleRequest', lmt_Debug, 'Retreiving executor ...', GetProcessID);

      _Executor := FLibrariesManager.getWorker(_ARequest.Lib);

      if _Executor = nil then  // Answer error: no executor found!
      begin
        LBLogger.Write(1, 'TFPWebProxy.DataModuleRequest', lmt_Warning, 'Executor not retrieved!', GetProcessID);
        _Answer := THTTPAnswer.Create;
        _Answer.Status := HTTP_AnswerCommandNotExecuted;
        _Answer.LastError := 'Generic worker not retrieved!';
        Self.SendAnswer(AResponse, _Answer);
        FreeAndNil(_Answer);
        FreeAndNil(_ARequest);
        Exit;
      end;

      LBLogger.Write(7, 'TFPWebProxy.DataModuleRequest', lmt_Debug, 'Executor retreived', GetProcessID);

      _Error := False;
      if Length(_ARequest.Token) > 0 then
      begin
        // If there is a token try to retrieve session data
        if GlobalTokenManager = nil then
          InitializeTokenManager();

        LBLogger.Write(5, 'TFPWebProxy.DataModuleRequest', lmt_Debug, 'Updating token and reading session data ...', GetProcessID);
        _Error := not GlobalTokenManager.UpdateToken(nil, _ARequest.Token, _SessionData);
        if _Error then
          _sError := GlobalTokenManager.LastError;
      end
      else begin
        LBLogger.Write(5, 'TFPWebProxy.DataModuleRequest', lmt_Debug, 'Token not found ...', GetProcessID);
        _Error := not _Executor.CanManageLogin();
        if _Error then
          _sError := 'Invalid token!'
        else
          LBLogger.Write(5, 'TFPWebProxy.DataModuleRequest', lmt_Debug, 'Library can be executed', GetProcessID);
      end;

      if not _Error then
      begin
        // Elaborating answer by external library
        _pAdditionalData := nil;
        _AdditionalDataLen := Length(_AdditionalData);
        if _AdditionalDataLen > 0 then
          _pAdditionalData := pByte(@_AdditionalData[1]);

        if _Executor.Execute(PChar(_ARequest.Request), PChar(_SessionData), _pAdditionalData, Length(_AdditionalData)) then
        begin
          if (Length(_ARequest.Token) = 0) then
          begin
            _Token := strpas(_Executor.getToken());
            if Length(_Token) > 0 then
            begin
              if GlobalTokenManager = nil then
                InitializeTokenManager();

               GlobalTokenManager.AddToken(_Token, strpas(_Executor.getSessionData()));
            end;
          end;
        end;
        AResponse.ContentType := strpas(_Executor.getResponseContentType());

        _Executor.getAnswerData(_Data, _DataLen);

        AResponse.CustomHeaders.AddText(_Executor.getResponseAdditionalHeaders());

        if (_Data <> nil) and (_DataLen > 0) then
        begin
          LBLogger.Write(5, 'TFPWebProxy.DataModuleRequest', lmt_Debug, PChar(Format('Data to send: %d', [_DataLen])), GetProcessID);

          _Response := TMemoryStream.Create;
          _Response.WriteBuffer(_Data^, _DataLen);
          AResponse.ContentStream := _Response;
          AResponse.FreeContentStream := True;
        end;

        AResponse.SendContent;
      end
      else begin         // Answer error
        _Answer := THTTPAnswer.Create;
        _Answer.Status := HTTP_AnswerStatusError;
        _Answer.LastError := _sError;
        Self.SendAnswer(AResponse, _Answer);
        FreeAndNil(_Answer);
      end;
    end;
    FreeAndNil(_ARequest);

  except
    on E: Exception do
      LBLogger.Write(1, 'TFPWebProxy.DataModuleRequest', lmt_Error, PChar(E.Message), GetProcessID);
  end;

  if _ARequest <> nil then
    _ARequest.Free;

  if _Answer <> nil then
    _Answer.Free;


  LBLogger.Write(5, 'TFPWebProxy.DataModuleRequest', lmt_Debug, '------> Request handled <------', GetProcessID);
end;

procedure TFPWebProxy.SendAnswer(AResponse: TResponse; AnAnswer: THTTPAnswer);
var
  _DataToSend : String;
  _jLen : Integer;
  _Response : TMemoryStream = nil;

begin
  try
    _DataToSend := AnAnswer.SerializeAsJSON;
    _jLen := Length(_DataToSend);

    AResponse.ContentType := HTTP_ContentType_JSON;
    AResponse.CustomHeaders.Values['JSONLEN'] := IntToStr(_jLen);

    if _jLen > 0 then
    begin

      _Response := TMemoryStream.Create;
      _Response.WriteBuffer(Pointer(@_DataToSend[1])^, _jLen);

      AResponse.ContentStream := _Response;
      _Response := nil;
      AResponse.FreeContentStream := True;
    end;

    AResponse.SendResponse;
    LBLogger.Write(7, 'TFPWebProxy.SendAnswer', lmt_Debug, 'Answer sent!', GetProcessID);

  except
    on E: Exception do
      LBLogger.Write(1, 'TFPWebProxy.SendAnswer', lmt_Error, PChar(E.Message), GetProcessID);
  end;

  if _Response <> nil then
    _Response.Free;
end;



procedure TFPWebProxy.getDataSent(ARequest: TRequest; out JSONData: String; out AdditionalData: String);
var
  _sJLen             : String;
  _JLen              : Integer;
  _AdditionalDataLen : Integer;


// ---------------------------------------------------------------------------------
// http://server/Commands?JLen=20    <!-- the data are posted and the json string lenght = 20
// http://server/Commands?JSON={}    <!-- the data are sent as query field value
//
//
// Posted data structure:
// [ JSON String with length = JLEN ][ RAW DATA .... Size = Request.ContentLength - JLEN ]

begin
  JSONData := '';
  AdditionalData := '';

  _sJLen := ARequest.QueryFields.Values['JLen'];
  if Length(_sJLen) = 0 then
  begin
    JSONData := ARequest.QueryFields.Values['JSON'];
    if Length(JSONData) > 0 then
      LBLogger.Write(3, 'TFPWebProxy.getDataSent', lmt_Debug, 'JSON data by get method', GetProcessID);
  end
  else begin
    _JLen := StrToIntDef(_sJLen, 0);
    if (_JLen > 0) and (_JLen < ARequest.ContentLength) then
    begin
      JSONData := Copy(ARequest.Content, 1, _JLen);
      _AdditionalDataLen := ARequest.ContentLength - _JLen;
      LBLogger.Write(3, 'TFPWebProxy.getDataSent', lmt_Debug, PChar(Format('Additional data size: %d',[_AdditionalDataLen])), GetProcessID);
      AdditionalData := Copy(ARequest.Content, _JLen + 1, _AdditionalDataLen);
    end
    else begin
      JSONData := ARequest.Content;
      AdditionalData := '';
    end;
  end;
end;


initialization
  RegisterHTTPModule('fpcProxyModule', TFPWebProxy);

end.

