unit uGlobals;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, ULBLogger, uTokenManager, uLibrariesManager, uLBDBConnectionManager;

type
  { TConfigData }

  TConfigData = class(TObject)
    strict private
      FConfigFolder : String;
      FSQLConnectionManager : TSQLConnectionManager;
      procedure setLogLevel(ConfigurationFile: String);
      procedure set_ConfigFolder(AValue: String);
    public
      constructor Create;
      destructor Destroy; override;

      property ConfigFolder: String read FConfigFolder write set_ConfigFolder;
      property SQLConnectionManager: TSQLConnectionManager read FSQLConnectionManager;
  end;

var
  GlobalLogFolder        : String = '';
  GlobalConfigFolder     : String = '';

  GlobalLogger           : ILBLogger = nil;
  GlobalConfigData       : TConfigData = nil;
  GlobalTokenManager     : TTokenManager = nil;
  GlobalLibrariesManager : TLibrariesManager = nil;


  // Ordine corretto con cui chiamare le funzioni di inizializzazione
  function InitializeFolders(): Boolean;
  function InitializeLogger(): Boolean;
  function InitializeConfigurationData(): Boolean;
  function InitializeTokenManager(): Boolean;
  function InitializeLibrariesManager(): Boolean; //TLibrariesManager; //Boolean;
  function DestroyGlobals(): Boolean;

implementation

uses
  uLBUtils, DOM, XMLRead;

const
  cConfigFileName   = String ('HttpdProxy.conf');
  cLibrariesMapFile = String ('HttpdProxyLibraries.dat');

function InitializeFolders(): Boolean;
const
  cSysTmpDir = String('/tmp/');
  {$ifdef Linux}
  cConfigFolder = String('/etc/LBSoftware/');
  {$else}
  cConfigFolder = String('c:\LBSoftware\');
  {$endif}

begin
  GlobalLogFolder := uLBUtils.getTemporaryFolder();

  if Length(GlobalLogFolder) = 0 then
  begin
    if DirectoryExists(cSysTmpDir) then
      GlobalLogFolder := cSysTmpDir;
  end;

  GlobalConfigFolder := cConfigFolder;

  Result := (Length(GlobalConfigFolder) > 0) and
            (Length(GlobalLogFolder) > 0);
end;

function InitializeLogger(): Boolean;
begin
  Result := False;

  if LBLogger = nil then
  begin
    LBLogger := TLBLogger.Create;
    LBLogger.FilePath := GlobalLogFolder;
    LBLogger.FileName := 'HttpdProxy.log';
    LBLogger.Write(1, 'fpcProxy.InitializeLogger', lmt_Info, '************  Logger created  ************', GetProcessID);

    GlobalLogger := LBLogger as ILBLogger;

    Result := True;
  end;
end;

function InitializeConfigurationData: Boolean;
begin
  Result := False;

  if GlobalConfigData = nil then
  begin
    LBLogger.Write(3, 'fpcProxy.InitializeConfigurationData', lmt_Debug, PChar(Format('Reading configuration data in folder <%s> ...', [GlobalConfigFolder])), GetProcessID);
    GlobalConfigData := TConfigData.Create;
    GlobalConfigData.ConfigFolder := GlobalConfigFolder;

    Result := True;
  end;
end;

function InitializeTokenManager: Boolean;
begin
  Result := False;

  if GlobalTokenManager = nil then
  begin
    LBLogger.Write(5, 'fpcProxy.InitializeTokenManager', lmt_Debug, 'Initializing TokenManager ...', GetProcessID);

    GlobalTokenManager := TTokenManager.Create;
    GlobalTokenManager.ConnectionManager := GlobalConfigData.SQLConnectionManager;

    LBLogger.Write(5, 'fpcProxy.InitializeTokenManager', lmt_Debug, 'TokenManager initialized', GetProcessID);

    Result := True;
  end;
end;

function InitializeLibrariesManager: Boolean;
begin
  Result := False;

  if GlobalLibrariesManager = nil then
  begin
    LBLogger.Write(5, 'InitializeLibrariesManager', lmt_Debug, 'Initializing libraries manager ...', GetProcessID);

    GlobalLibrariesManager := TLibrariesManager.Create;
    GlobalLibrariesManager.ConfigFolder := GlobalConfigData.ConfigFolder;
    GlobalLibrariesManager.FileName := cLibrariesMapFile;
    Result := True;

    LBLogger.Write(5,'InitializeLibrariesManager', lmt_Debug, 'Libraries manager initialized', GetProcessID);
  end;
end;

function DestroyGlobals: Boolean;
begin
  Result := False;

  try
    LBLogger.Write(5, 'DestroyGlobals', lmt_Debug, 'Destroying global variables ...', GetProcessID);
    Sleep(200);

    if GlobalTokenManager <> nil then
      FreeAndNil(GlobalTokenManager);

    if GlobalLibrariesManager <> nil then
      FreeAndNil(GlobalLibrariesManager);

    if GlobalConfigData <> nil then
      FreeAndNil(GlobalConfigData);

    LBLogger := nil;
    GlobalLogger := nil;

    Result := True;

  except

  end;
end;


{ TConfigData }

procedure TConfigData.setLogLevel(ConfigurationFile: String);
var
  _Doc : TXMLDocument = nil;
  _Root : TDOMNode;
  _Item : TDOMNode;
  _Level : Integer;

const
  cLogLevelNodeName = WideString('LogLevel');

begin
  try

    if FileExists(ConfigurationFile) then
    begin
      ReadXMLFile(_Doc, ConfigurationFile);
      if _Doc <> nil then
      begin
        _Root := _Doc.DocumentElement;

        if _Root <> nil then
        begin
          _Item := _Root.FindNode(cLogLevelNodeName);
          if _Item <> nil then
          begin
            _Level := StrToIntDef(AnsiString(_Item.TextContent), 0);
            if _Level > 0 then
              LBLogger.MaxLogLevel := _Level;
          end;
        end;
      end
      else
        LBLogger.Write(1, 'TConfigData.setLogLevel', lmt_Warning, PChar(Format('Not enabled to read <%s>', [ConfigurationFile])), GetProcessID);
    end
    else
      LBLogger.Write(1, 'TConfigData.setLogLevel', lmt_Warning, PChar(Format('Configuration file <%s> not found!', [ConfigurationFile])), GetProcessID);

  except
    on E: Exception do
      LBLogger.Write(1, 'TConfigData.setLogLevel', lmt_Error, PChar(E.Message), GetProcessID);
  end;

  if _Doc <> nil then
    _Doc.Free;
end;

procedure TConfigData.set_ConfigFolder(AValue: String);
begin
  FConfigFolder := IncludeTrailingPathDelimiter(AValue);

  FSQLConnectionManager.LoadConfiguration(FConfigFolder + cConfigFileName);

  Self.SetLogLevel(FConfigFolder + cConfigFileName);
end;

constructor TConfigData.Create;
begin
  inherited Create;
  FSQLConnectionManager := TSQLConnectionManager.Create;
end;

destructor TConfigData.Destroy;
begin
  FreeAndNil(FSQLConnectionManager);
  inherited Destroy;
end;


end.

