unit uLibrariesManager;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, dynlibs, uCommonTypes, ULBLogger;

type
  TgetWorkerObject = function(out AWorker: IGenericWorker): boolean; stdcall;

  TInitializeSO = function(pConfigFilePath: pAnsiChar; ALogger: ILBLogger): Boolean; stdcall;

  TFinalizeSO = function(): Boolean; stdcall;


const
  cGetWorkerObjectFunctionName = AnsiString('getWorker');
  cFinalizeSO                  = AnsiString('Finalize');
  cInitializeSO                = AnsiString('Initialize');

type

  { TLibraryData }

  TLibraryData = class(TObject)
    strict private
      FHandle: TLibHandle;
      FInitialized : Boolean;
      FLibName : AnsiString;
      function get_Linked: Boolean;
      function UnloadLibrary: Boolean;
    public
      constructor Create(LibName: AnsiString; ConfigurationPath: AnsiString);
      destructor Destroy; override;

      function getWorker(): IGenericWorker;

      property Linked: Boolean read get_Linked;
  end;

  { TLibrariesManager }

  TLibrariesManager = class(TObject)
    strict private
      FConfigFolder : String;
      FFileName : String;
      FLibraries : TStringList;

      function LoadLibrary(LibId: String): Integer;
      procedure set_ConfigFolder(AValue: String);
      procedure set_FileName(AValue: String);
    public
      constructor Create;
      destructor Destroy; override;

      function getWorker(LibId: String): IGenericWorker;

      property ConfigFolder: String write set_ConfigFolder;
      property FileName: String write set_FileName;
  end;

implementation

{ TLibrariesManager }

function TLibrariesManager.LoadLibrary(LibId: String): Integer;
var
  _FileName : String;
  _ProxyFile : TStringList;
  _libFileName : String;
  _Lib : TLibraryData;

begin
  Result := -1;

  try
    LBLogger.Write(5, 'TLibrariesManager.LoadLibrary', lmt_Debug, PChar(Format('Loading library <%s> ...', [LibId])), GetProcessID);

    if (Length(FFileName) > 0) and (Length(FConfigFolder) > 0) then
    begin
      _FileName := FConfigFolder + FFileName;  // In this file it finds which is the file for the library requested
      if not FileExists(_FileName) then
      begin
        LBLogger.Write(1, 'TLibrariesManager.LoadLibrary', lmt_Warning, PChar(Format('Configuration file <%s> not found!', [_FileName])), GetProcessID);
        Exit;
      end;

      _ProxyFile := TStringList.Create;
      _ProxyFile.LoadFromFile(_FileName); // The file has n lines that are structured as key = Value
      _ProxyFile.NameValueSeparator := '=';
      _libFileName := _ProxyFile.Values[LibId];

      FreeAndNil(_ProxyFile);
      if Length(_libFileName) = 0 then
      begin
        LBLogger.Write(1, 'TLibrariesManager.LoadLibrary', lmt_Warning, PChar(Format('No file for library <%s>', [LibId])), GetProcessID);
        Exit;
      end;

      if not FileExists(_libFileName) then
      begin
        LBLogger.Write(1, 'TLibrariesManager.LoadLibrary', lmt_Warning, PChar(Format('File <%s> for library <%s> doesn''t exist!', [_libFileName, LibId])), GetProcessID);
        Exit;
      end;

      _Lib := TLibraryData.Create(_libFileName, FConfigFolder);

      if not _Lib.Linked then
      begin
        LBLogger.Write(1, 'TLibrariesManager.LoadLibrary', lmt_Warning, PChar(Format('Library <%s> not linked!', [_libFileName])), GetProcessID);
        FreeAndNil(_Lib);
        Exit;
      end;

      Result := FLibraries.AddObject(LibId, _Lib);
    end
    else
      LBLogger.Write(1, 'TLibrariesManager.LoadLibrary', lmt_Debug, PChar(Format('Configuration file not set! Path: <%s>  File: <%s>', [FFileName, FConfigFolder])), GetProcessID);

  except
    on E: Exception do
    begin
      Result := -1;
      LBLogger.Write(1, 'TLibrariesManager.LoadLibrary', lmt_Error, PChar(E.Message), GetProcessID);
    end;
  end;
end;

procedure TLibrariesManager.set_ConfigFolder(AValue: String);
begin
  FConfigFolder := AValue;
  if Length(FConfigFolder) > 0 then
  begin
    if RightStr(FConfigFolder, 1) <> PathDelim then
      FConfigFolder += PathDelim;
  end;
  LBLogger.Write(5, 'TLibrariesManager.set_ConfigFolder', lmt_Debug, PChar(Format('Configuration folder: <%s>', [FConfigFolder])), GetProcessID);
end;

procedure TLibrariesManager.set_FileName(AValue: String);
begin
  FFileName := AValue;
  LBLogger.Write(5, 'TLibrariesManager.set_FileName', lmt_Debug, PChar(Format('Configuration file: <%s>', [FFileName])), GetProcessID);
end;

constructor TLibrariesManager.Create;
begin
  FConfigFolder := '';
  FFileName := '';
  FLibraries := TStringList.Create;
  FLibraries.Duplicates := dupError;
  FLibraries.CaseSensitive := True;
  FLibraries.Sorted := True;
  FLibraries.OwnsObjects := True;
end;

destructor TLibrariesManager.Destroy;
begin
  LBLogger.Write(5, 'TLibrariesManager.Destroy', lmt_Debug, PChar('Unloading libraries ...'), GetProcessID);

  try
    FLibraries.Free;
  except
    on E: Exception do
      LBLogger.Write(5, 'TLibrariesManager.Destroy', lmt_Error, PChar(E.Message), GetProcessID);
  end;

  inherited Destroy;
end;

function TLibrariesManager.getWorker(LibId: String): IGenericWorker;
var
  _idx : Integer;
  _Lib : TLibraryData = nil;

begin
  Result := nil;

  if Self = nil then
    Exit;

  if Length(LibId) = 0 then
  begin
    LBLogger.Write(1, 'TLibrariesManager.getGenericWorker', lmt_Warning, 'No library name!', GetProcessID);
    Exit;
  end;

  LibId := LowerCase(LibId);

  _idx := FLibraries.IndexOf(LibId);
  if _idx = -1 then
  begin
    LBLogger.Write(5, 'TLibrariesManager.getGenericWorker', lmt_Debug, PChar(Format('Library <%s> not yet loaded ...', [LibId])), GetProcessID);
    _idx := Self.LoadLibrary(LibId);
  end;

  if _idx > -1 then
  begin
    _Lib := FLibraries.Objects[_idx] as TLibraryData;
    Result := _Lib.getWorker();
  end
  else
    LBLogger.Write(1, 'TLibrariesManager.getGenericWorker', lmt_Warning, 'Library not found and not loaded', GetProcessID);

end;


{ TLibraryData }

function TLibraryData.get_Linked: Boolean;
begin
  Result := (FHandle <> 0) and
            FInitialized;
end;

function TLibraryData.UnloadLibrary: Boolean;
var
  _Finalize : TFinalizeSO;
begin
  Result := False;
  LBLogger.Write(5, 'TLibraryData.UnloadLibrary', lmt_Debug, PChar(Format('<%s> started',[FLibName])), GetProcessID);

  try
    if FHandle <> 0 then
    begin
      if FInitialized then
      begin
        _Finalize := TFinalizeSO(GetProcAddress(FHandle, cFinalizeSO));
        if _Finalize <> nil then
        begin
          LBLogger.Write(5, 'TLibraryData.UnloadLibrary', lmt_Debug, PChar(Format('Finalizing library <%s> ...',[FLibName])), GetProcessID);
          _Finalize();
        end
        else
          LBLogger.Write(1, 'TLibraryData.UnloadLibrary', lmt_Warning, PChar(Format('Finalize function not found in library <%s> ...', [FLibName])), GetProcessID);
      end
      else
        LBLogger.Write(5, 'TLibraryData.UnloadLibrary', lmt_Debug, PChar(Format('Library <%s> not initialized',[FLibName])), GetProcessID);

      FreeLibrary(FHandle);
      FHandle := 0;
    end
    else
        LBLogger.Write(5, 'TLibraryData.UnloadLibrary', lmt_Debug, PChar(Format('Library <%s> not loaded', [FLibName])), GetProcessID);

    FInitialized := False;
    Result := True;

  except
    on E: Exception do
      LBLogger.Write(1, 'TLibraryData.UnloadLibrary', lmt_Error, PChar(E.Message), GetProcessID);
  end;
end;

constructor TLibraryData.Create(LibName: AnsiString; ConfigurationPath: AnsiString);
var
  _Initialize         : TInitializeSO = nil;
  _pConfigurationPath : pAnsiChar = nil;

begin
  inherited Create;

  FHandle := 0;
  FInitialized := False;
  FLibName := LibName;

  LBLogger.Write(5, 'TLibraryData.Create', lmt_Debug, PChar(Format('Loading module <%s> ...', [LibName])), GetProcessID);

  if FileExists(LibName) then
  begin
    try
      FHandle := LoadLibrary(LibName);
      if FHandle <> 0 then
      begin
        LBLogger.Write(5, 'TLibraryData.Create', lmt_Debug, 'Library loaded', GetProcessID);

        _Initialize := TInitializeSO(GetProcAddress(FHandle, cInitializeSO));
        if _Initialize <> nil then
        begin
          LBLogger.Write(5, 'TLibraryData.Create', lmt_Debug, 'Initialize function retrieved', GetProcessID);

          if Length(ConfigurationPath) > 0 then
            _pConfigurationPath := @ConfigurationPath[1];

          FInitialized := _Initialize(_pConfigurationPath, LBLogger as ILBLogger);

          LBLogger.Write(5, 'TLibraryData.Create', lmt_Debug, 'Initialize function executed', GetProcessID);
        end
        else
          LBLogger.Write(1, 'TLibraryData.Create', lmt_Warning, PChar(Format('Error retrieving <%s> function!',[cInitializeSO])), GetProcessID);


        if not FInitialized then
        begin
          LBLogger.Write(1, 'TLibraryData.Create', lmt_Warning, 'Library not initialized!', GetProcessID);
          Self.UnloadLibrary();
        end;
      end
      else
        LBLogger.Write(1, 'TLibraryData.Create', lmt_Warning, PChar(Format('Error loading library <%s>!', [LibName])), GetProcessID);

    except
      on E: Exception do
      begin
        LBLogger.Write(1, 'TLibraryData.Create', lmt_Error, PChar(Format('Error: <%s>', [E.Message])), GetProcessID);
        Self.UnloadLibrary();
      end;
    end;
  end
  else
    LBLogger.Write(1, 'TLibraryData.Create', lmt_Warning, PChar(Format('Library <%s> not found!', [LibName])), GetProcessID);

  if FInitialized then
    LBLogger.Write(5, 'TLibraryData.Create', lmt_Debug, PChar(Format('Library <%s> loaded and initialized!', [LibName])), GetProcessID);
end;

destructor TLibraryData.Destroy;
begin
  try
    LBLogger.Write(5, 'TLibraryData.Destroy', lmt_Debug, 'Destroying ...', GetProcessID);

    Self.UnloadLibrary();
    inherited Destroy;

  except
    on E: Exception do
    begin
      LBLogger.Write(1, 'TLibraryData.Destroy', lmt_Error, PChar(E.Message), GetProcessID);
    end;
  end;
end;

function TLibraryData.getWorker: IGenericWorker;
var
  _getWorker : TgetWorkerObject;

begin
  Result := nil;
  if Self.Linked then
  begin
    _getWorker := TgetWorkerObject(GetProcAddress(FHandle, cGetWorkerObjectFunctionName));
    if _getWorker = nil then
      LBLogger.Write(1, 'TLibraryData.getWorker', lmt_Warning, PChar(Format('Error retrieving GenericWorker in library <%s>!',[FLibName])), GetProcessID)
    else
      _getWorker(Result);
  end
  else
    LBLogger.Write(1, 'TLibraryData.getWorker', lmt_Warning, PChar(Format('Library <%s> not initialized',[FLibName])));
end;

end.

