unit uTokenManager;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, ULBLogger, uLBDBConnectionManager, sqldb;

type

  { TTokenManager }
  TTokenManager = class(TObject)
    strict private
      FDBAlias : String;
      FTokenTableName : String;
      FTokenLifeTime : Double;
      FLastError : String;

      FSQLConnectionManager : TSQLConnectionManager;

      procedure set_SQLConnectionManager(AValue: TSQLConnectionManager);
      function VerifyConnection(AConnection: TSQLConnection): Boolean;
      procedure RollBack(AConnection: TSQLConnection);
      function getDBConnection(): TSQLConnection;

      procedure CreateTokensTable(AConnection: TSQLConnection; DBType: uLBDBConnectionManager.TDBType);
    public
      constructor Create;

      function UpdateToken(AConnection: TSQLConnection; AToken: AnsiString; var SessionData: AnsiString): Boolean;
      function AddToken(AToken: AnsiString; SessionData: AnsiString): Boolean;

      property ConnectionManager: TSQLConnectionManager write set_SQLConnectionManager;
      property LastError: String read FLastError;
  end;

implementation


const
  cDBTokenDateFormat = String('yyyymmddhhnnss');


{ TTokenManager }

function TTokenManager.VerifyConnection(AConnection: TSQLConnection): Boolean;
begin
  Result := False;
  if AConnection <> nil then
  begin
    if not AConnection.Connected then
    begin
      LBLogger.Write(5, 'TTokenManager.VerifyConnection', lmt_Info, 'DB connection closed, reconnecting', GetThreadID);
      try
        AConnection.Open;
        Result := True;
      except
        on E: Exception do
          LBLogger.Write(1, 'TTokenManager.VerifyConnection', lmt_Error, PChar(Format('Database NOT CONNECTED: %s', [E.Message])), GetThreadID);
      end;
    end
    else
      Result := True;
  end;
end;

procedure TTokenManager.set_SQLConnectionManager(AValue: TSQLConnectionManager);
var
  _AConnection : TSQLConnection;

begin
  FSQLConnectionManager := AValue;
  _AConnection := Self.getDBConnection();
  Self.CreateTokensTable(_AConnection, FSQLConnectionManager.getDatabaseType(FDBAlias));
  FSQLConnectionManager.CloseConnection(_AConnection);
end;

procedure TTokenManager.RollBack(AConnection: TSQLConnection);
begin
  if (AConnection <> nil) and
     AConnection.Connected and
     AConnection.Transaction.Active then
  begin
    AConnection.Transaction.Active := False;
    LBLogger.Write(3, 'TTokenManager.RollBack', lmt_Debug, 'Transaction rolled back!', GetProcessID);
  end;
end;

procedure TTokenManager.CreateTokensTable(AConnection: TSQLConnection; DBType: uLBDBConnectionManager.TDBType);
begin
  if (AConnection <> nil) then
  begin
    if DBType <> dbt_Unknown then
    begin
      try
        LBLogger.Write(3, 'TTokenManager.CreateTokensTable', lmt_Debug, PChar(Format('Creating <%s> table if not exists', [FTokenTableName])), GetProcessID);

        AConnection.Transaction.StartTransaction;
        case DBType of
          dbt_ODBC,
          dbt_MySQL55:
            begin
              AConnection.ExecuteDirect('CREATE TABLE IF NOT EXISTS ' + FTokenTableName + ' ' +
                                        '(Token VARCHAR(50) NOT NULL,' +
                                        'LastRequestTime VARCHAR(14) NOT NULL,' +
                                        'SessionData LONGTEXT DEFAULT NULL,' +
                                        'PRIMARY KEY (Token),' +
                                        'INDEX LastRequestIdx (LastRequestTime));');
            end;

          dbt_SQLite:
            begin
              AConnection.ExecuteDirect('CREATE TABLE IF NOT EXISTS ' + FTokenTableName + ' ' +
                                        '(Token CHAR(50) NOT NULL,' +
                                         'LastRequestTime CHAR(14) NOT NULL,' +
                                         'SessionData TEXT DEFAULT NULL,' +
                                         'PRIMARY KEY (Token ASC));');
              AConnection.ExecuteDirect('CREATE INDEX IF NOT EXISTS LastRequestTimeIdx ON ' + FTokenTableName + ' (LastRequestTime);');
            end;
        end;
        AConnection.Transaction.Commit();

      except
        on E: Exception do
        begin
          if AConnection.Transaction.Active then
            AConnection.Transaction.Rollback;

          LBLogger.Write(1, 'TTokenManager.CreateTokensTable', lmt_Error, PChar(E.Message), GetProcessID);
        end;
      end;
    end
    else
      LBLogger.Write(1, 'TTokenManager.CreateTokensTable', lmt_Warning, 'Unknown DB type!', GetProcessID);

  end
  else
    LBLogger.Write(1, 'TTokenManager.CreateTokensTable', lmt_Warning, 'DBs Connection not set!', GetProcessID);

end;

function TTokenManager.getDBConnection: TSQLConnection;
begin
  Result := nil;

  if (FSQLConnectionManager <> nil) then
  begin
    LBLogger.Write(5, 'TTokenManager.getDBConnection', lmt_Debug, 'Getting new token connection ...', GetProcessID);
    Result := FSQLConnectionManager.getConnection(FDBAlias);
  end;
end;

constructor TTokenManager.Create;
begin
  inherited Create;
  FSQLConnectionManager := nil;
  FDBAlias := 'DBTokens';
  FTokenTableName := 'Tokens';

  // Expiration time = 9 hours of inactivity
  FTokenLifeTime := (9 * 3600) / SecsPerDay;
end;

function TTokenManager.UpdateToken(AConnection: TSQLConnection; AToken: AnsiString; var SessionData: AnsiString): Boolean;
var
  _SQLQuery    : TSQLQuery = nil;
  _DestroyConn : Boolean = False;
  _AConnection : TSQLConnection = nil;

begin
  Result := False;
  SessionData := '';

  if (Length(AToken) > 0) then
  begin
    try
      if AConnection = nil then
      begin
        _AConnection := Self.getDBConnection();
        _DestroyConn := True;
      end
      else
        _AConnection := AConnection;

      if _AConnection <> nil then
      begin
        _AConnection.StartTransaction;

        _SQLQuery := TSQLQuery.Create(nil);
        _SQLQuery.DataBase := _AConnection;

        _SQLQuery.SQL.Text := 'SELECT SessionData FROM ' + FTokenTableName + ' WHERE (Token = :Tk);';
        _SQLQuery.Params.ParamByName('Tk').AsString := AToken;

        _SQLQuery.Open();
        if not (_SQLQuery.EOF and _SQLQuery.BOF) then
        begin
          _SQLQuery.First;
          SessionData := _SQLQuery.Fields[0].AsString;
          Result := True;
        end
        else begin
          FLastError := 'Token not found!';
          LBLogger.Write(1, 'TTokenManager.UpdateToken', lmt_Warning, PChar(Format('Token: %s  -  Session data not found!',[AToken])), GetProcessID);
        end;

        _SQLQuery.Close();

        if Result then
        begin
          _SQLQuery.Params.Clear();
          _SQLQuery.SQL.Text := 'UPDATE ' + FTokenTableName + ' SET LastRequestTime = :ATime WHERE (Token = :Tk);';
          _SQLQuery.Params.ParamByName('Tk').AsString := AToken;
          _SQLQuery.Params.ParamByName('ATime').AsString := FormatDateTime(cDBTokenDateFormat, Now());
          _SQLQuery.ExecSQL();
          _SQLQuery.Close();
        end;

        _SQLQuery.Params.Clear;
        _SQLQuery.SQL.Text := 'DELETE FROM Tokens WHERE (LastRequestTime < :LifeTime);';
        _SQLQuery.Params.ParamByName('LifeTime').AsString := FormatDateTime(cDBTokenDateFormat, Now() - FTokenLifeTime);
        _SQLQuery.ExecSQL();
        _SQLQuery.Close();

        _AConnection.Transaction.Commit;
      end
      else
        LBLogger.Write(1, 'TTokenManager.UpdateToken', lmt_Warning, 'No database connection!', GetProcessID);

    except
      on E: Exception do
      begin
        Result := False;
        FLastError := E.Message;
        LBLogger.Write(1, 'TTokenManager.UpdateToken', lmt_Error, PChar(E.Message), GetProcessID);
        RollBack(_AConnection);
      end;
    end;

    if _SQLQuery <> nil then
      FreeAndNil(_SQLQuery);

    if (_AConnection <> nil) and _DestroyConn then
      FSQLConnectionManager.CloseConnection(_AConnection);

  end
  else
    LBLogger.Write(1, 'TTokenManager.UpdateToken', lmt_Warning, 'Token not set!', GetProcessID);
end;

function TTokenManager.AddToken(AToken: AnsiString; SessionData: AnsiString): Boolean;
var
  _SQLQuery          : TSQLQuery = nil;
  _SessionData       : AnsiString = '';
  _DestroyConnection : Boolean = false;
  _AConnection       : TSQLConnection;

begin
  Result := False;
  if (Length(AToken) > 0) then
  begin
    if (FSQLConnectionManager <> nil) then
    begin
       _AConnection := Self.getDBConnection();
       _DestroyConnection := True;
    end;

    if _AConnection <> nil then
    begin
      Result := Self.UpdateToken(_AConnection, AToken, _SessionData);
      if not Result then
      begin
        try
          LBLogger.Write(5, 'TTokenManager.AddToken', lmt_Debug, PChar(Format('Adding token: <%s>  - SessionData: <%s>',[AToken, SessionData])), GetProcessID);

          if not _AConnection.Transaction.Active then
          begin
            LBLogger.Write(5, 'TTokenManager.AddToken', lmt_Debug, 'Starting transaction', GetProcessID);
            _AConnection.StartTransaction;
          end
          else
            LBLogger.Write(3, 'TTokenManager.AddToken', lmt_Debug, 'Transaction already active', GetProcessID);


          _SQLQuery := TSQLQuery.Create(nil);
          _SQLQuery.DataBase := _AConnection;

          _SQLQuery.SQL.Text := 'INSERT INTO ' + FTokenTableName + ' (Token, LastRequestTime, SessionData) ' +
                                'VALUES (:Tkn, :ATime, :SessionD);';
          _SQLQuery.Params.ParamByName('Tkn').AsString := AToken;
          _SQLQuery.Params.ParamByName('SessionD').AsString := SessionData;
          _SQLQuery.Params.ParamByName('ATime').AsString := FormatDateTime(cDBTokenDateFormat, Now());
          _SQLQuery.ExecSQL();
          _SQLQuery.Close;

          _AConnection.Transaction.Commit;
          Result := True;

        except
          on E: Exception do
          begin
            Result := False;
            RollBack(_AConnection);
            LBLogger.Write(1, 'TTokenManager.AddToken', lmt_Error, PChar(E.Message), GetProcessID);
          end;
        end;

        if _SQLQuery <> nil then
          FreeAndNil(_SQLQuery);
      end;
      if _DestroyConnection and (_AConnection <> nil) then
        FSQLConnectionManager.CloseConnection(_AConnection);
    end
    else
      LBLogger.Write(1, 'TTokenManager.AddToken', lmt_Warning, 'No database connection!', GetProcessID);

  end
  else
    LBLogger.Write(1, 'TTokenManager.AddToken', lmt_Warning, 'Token not found or connection not set!', GetProcessID);

end;

end.

