unit uLBDBConnectionManager;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, contnrs, sqldb, ULBLogger, DOM;

type
  TDBType = (dbt_Unknown = 0,
             dbt_SQLite  = 1,
             dbt_ODBC    = 2,
             dbt_MySQL55 = 3);


  TNotifyNewDBEvent = procedure (AConnection: TSQLConnection; DBType: TDBType) of Object;

  { TSQLDBInfo }

  TSQLDBInfo = class(TObject)
    strict private
      FOnNewDatabase : TNotifyNewDBEvent;

      FHost          : String;
      FPort          : Word;
      FUser          : String;
      FPassWord      : String;
      FDBName        : String;
      FDBType        : TDBType;
      FODBC          : String;

      function get_DataComplete: Boolean;
      procedure DoNewDatabase(AConnection: TSQLConnection);
    public
      constructor Create;

      function LoadConfiguration(ANode: TDOMNode): Boolean;
      function getConnection(): TSQLConnection;

      property Host     : String  read FHost;
      property Port     : Word    read FPort;
      property User     : String  read FUser;
      property PassWord : String  read FPassWord;
      property DBName   : String  read FDBName;
      property DBType   : TDBType read FDBType;
      property ODBC     : String  read FODBC;

      property DataComplete: Boolean read get_DataComplete;

      property OnNewDatabase : TNotifyNewDBEvent read FOnNewDatabase write FOnNewDatabase;
  end;



  { TSQLConnectionManager }

  TSQLConnectionManager = class(TObject)
    strict private
      FOnNewDatabase : TNotifyNewDBEvent;
      FConnectionsInfo : TStringList;
      FActiveConnections : TObjectList;
    public
      constructor Create;
      destructor Destroy; override;

      function LoadConfiguration(AnXMLFile: String): Boolean;
      function LoadConfigurationFromXML(ParentNode: TDOMNode): boolean;

      function getConnection(DBAlias: String): TSQLConnection;
      function getDatabaseType(DBAlias: String): TDBType;

      procedure CloseConnection(var AConnection : TSQLConnection);

      property OnNewDatabase: TNotifyNewDBEvent read FOnNewDatabase write FOnNewDatabase;
  end;


  function getMySQLLastInsertID(AQuery: TSQLQuery): Integer;

implementation

uses
  mysql55conn, sqlite3conn, odbcconn, sqlite3dyn, XMLRead;



function getMySQLLastInsertID(AQuery: TSQLQuery): Integer;
begin
  Result := 0;
  if AQuery <> nil then
  begin
    AQuery.SQL.Text := 'SELECT LAST_INSERT_ID();';
    AQuery.Open();
    if not (AQuery.BOF and AQuery.EOF) then
      Result := AQuery.Fields[0].AsInteger;
    AQuery.Close();
  end;
end;


{ TSQLConnectionManager }

function TSQLConnectionManager.getConnection(DBAlias: String): TSQLConnection;
var
  _idx : Integer;
  _ConnInfo : TSQLDBInfo;

begin
  Result := nil;

  try

    if Length(DBAlias) = 0 then
    begin
      LBLogger.Write(1, 'TSQLConnectionManager.getConnection', lmt_Warning, 'No DBAlias!', GetProcessID);
      Exit;
    end;

    _idx := FConnectionsInfo.IndexOf(DBAlias);
    if _idx > -1 then
    begin
      _ConnInfo := FConnectionsInfo.Objects[_idx] as TSQLDBInfo;
      Result := _ConnInfo.getConnection();
      if Result <> nil then
        FActiveConnections.Add(Result);

    end
    else
      LBLogger.Write(1, 'TSQLConnectionManager.getConnection', lmt_Warning, 'DBAlias not found!', GetProcessID);

  except
    on E: Exception do
      LBLogger.Write(1, 'TSQLConnectionManager.getConnection', lmt_Error, PChar(E.Message), GetProcessID);
  end;
end;

function TSQLConnectionManager.getDatabaseType(DBAlias: String): TDBType;
var
  _idx : Integer;
  _ConnInfo : TSQLDBInfo;

begin
  Result := dbt_Unknown;
  if Length(DBAlias) > 0 then
  begin
    _idx := FConnectionsInfo.IndexOf(DBAlias);
    if _idx > -1 then
    begin
      _ConnInfo := FConnectionsInfo.Objects[_idx] as TSQLDBInfo;
      Result := _ConnInfo.DBType;
    end;
  end;
end;

constructor TSQLConnectionManager.Create;
begin
  inherited Create;

  FConnectionsInfo := TStringList.Create;
  FConnectionsInfo.OwnsObjects := True;
  FConnectionsInfo.Sorted := True;
  FConnectionsInfo.Duplicates := dupIgnore;

  FActiveConnections := TObjectList.Create(False);
end;

destructor TSQLConnectionManager.Destroy;
var
  i : Integer;
  _Conn : TSQLConnection;

begin
  if FActiveConnections.Count > 0 then
  begin
    for i := 0 to FActiveConnections.Count - 1 do
    begin
      _Conn := FActiveConnections.Items[0] as TSQLConnection;
      Self.CloseConnection(_Conn);
    end;
  end;

  FActiveConnections.Free;
  FConnectionsInfo.Free;

  inherited Destroy;
end;

function TSQLConnectionManager.LoadConfiguration(AnXMLFile: String): Boolean;
var
  _Doc : TXMLDocument = nil;

begin
  Result := False;

  try

    if FileExists(AnXMLFile) then
    begin
      ReadXMLFile(_Doc, AnXMLFile);
      if _Doc <> nil then
        Result := Self.LoadConfigurationFromXML(_Doc.DocumentElement);
    end
    else
      LBLogger.Write(1, 'TSQLConnectionManager.LoadConfiguration', lmt_Warning, PChar(Format('Configuration file <%s> not found!', [AnXMLFile])), GetProcessID);


  except
    on E: Exception do
      LBLogger.Write(1, 'TSQLConnectionManager.LoadConfiguration', lmt_Error, PChar(E.Message), GetProcessID);
  end;

  if _Doc <> nil then
    _Doc.Free;
end;


function TSQLConnectionManager.LoadConfigurationFromXML(ParentNode: TDOMNode): boolean;
var
  _ConnectionsNode : TDOMNode;
  _SingleConnNode : TDomNode;
  _Item : TDOMNode;
  _AConnInfo : TSQLDBInfo;
  _Alias : String;
  i : Integer;

const
  cDBConnections = WideString('DBConnections');
  cDBSingleConn  = WideString('Connection');
  cDBAlias       = WideString('DBAlias');

// ----------------------------------------------------------------------------------
// <DBConnections>
//    <Connection>
//        <DBType></DBType>
//        <DBHost></DBHost>
//        <DBPort></DBPort>
//        <DBPassword></DBPassword>
//        <DBUser></DBUser>
//        <DBName></DBName>
//        <DBODBC></DBODBC>
//        <DBAlias></DBAlias>
//    </Connection>
// </DBConnections>
// ----------------------------------------------------------------------------------

begin
  Result := False;

  if ParentNode = nil then
  begin
    LBLogger.Write(1, 'TSQLConnectionManager.LoadConfigurationFromXML', lmt_Warning, 'No parent node!', GetProcessID);
    Exit;
  end;

  _ConnectionsNode := ParentNode.FindNode(cDBConnections);
  if _ConnectionsNode = nil then
  begin
    LBLogger.Write(1, 'TSQLConnectionManager.LoadConfigurationFromXML', lmt_Warning, '<DBConnections> node not found!', GetProcessID);
    Exit;
  end;

  FConnectionsInfo.Clear;

  if _ConnectionsNode.ChildNodes.Count > 0 then
  begin
    for i := 0 to _ConnectionsNode.ChildNodes.Count - 1 do
    begin
      _SingleConnNode := _ConnectionsNode.ChildNodes.Item[i];
      if _SingleConnNode.NodeName = cDBSingleConn then
      begin
        _Item := _SingleConnNode.FindNode(cDBAlias);
        if _Item <> nil then
          _Alias := LowerCase(AnsiString(_Item.TextContent))
        else
          _Alias := '';

        if Length(_Alias) = 0 then
        begin
          LBLogger.Write(1, 'TSQLConnectionManager.LoadConfigurationFromXML', lmt_Warning, PChar(Format('Item %d: alias not found!', [i])), GetProcessID);
          Continue;
        end;

        _AConnInfo := TSQLDBInfo.Create;
        _AConnInfo.OnNewDatabase := FOnNewDatabase;
        if _AConnInfo.LoadConfiguration(_SingleConnNode) then
          FConnectionsInfo.AddObject(_Alias, _AConnInfo)
        else begin
          LBLogger.Write(1, 'TSQLConnectionManager.LoadConfigurationFromXML', lmt_Warning, PChar(Format('Item %d: error reading configuration!', [i])), GetProcessID);
          _AConnInfo.Free;
        end;
      end;
    end;

    Result := FConnectionsInfo.Count > 0;
  end;
end;


procedure TSQLConnectionManager.CloseConnection(var AConnection: TSQLConnection);
var
  _Transaction : TSQLTransaction;

begin
  if AConnection <> nil then
  begin

    try
      FActiveConnections.Extract(AConnection);
      try
        if AConnection.Connected then
        begin
          AConnection.Transaction.Rollback;
          AConnection.Close;
        end;
      except
        on E: Exception do
          LBLogger.Write(1, 'TSQLConnectionManager.CloseConnection', lmt_Error, PChar(Format('Error closing database connection: %s',[E.Message])), GetProcessID);
      end;

      _Transaction := AConnection.Transaction;
      FreeAndNil(AConnection);
      _Transaction.Free;

      LBLogger.Write(5, 'TSQLConnectionManager.CloseConnection', lmt_Debug, 'Database connection closed!', GetProcessID);
    except
      on E: Exception do
        LBLogger.Write(1, 'TSQLConnectionManager.CloseConnection', lmt_Error, PChar(E.Message), GetProcessID);
    end;

  end;
end;

{ TSQLDBInfo }

function TSQLDBInfo.get_DataComplete: Boolean;
begin
  Result := False;

  case FDBType of
    dbt_Unknown : ;

    dbt_MySQL55 : Result := (Length(FHost) > 0) and
                            (FPort > 0) and
                            (Length(FUser) > 0) and
                            (Length(FDBName) > 0);

    dbt_SQLite  : Result := Length(FDBName) > 0;

    dbt_ODBC    : Result := Length(FODBC) > 0;
  end;
end;

procedure TSQLDBInfo.DoNewDatabase(AConnection: TSQLConnection);
begin
  if Assigned(FOnNewDatabase) then
    FOnNewDatabase(AConnection, FDBType);
end;

constructor TSQLDBInfo.Create;
begin
  inherited Create;
  FHost     := '';
  FPort     := 0;
  FUser     := '';
  FPassWord := '';
  FDBName   := '';
  FDBType   := dbt_Unknown;
end;

function TSQLDBInfo.LoadConfiguration(ANode: TDOMNode): Boolean;
var
  _Item : TDOMNode;

const
  cDBType        = WideString('DBType');
  cHost          = WideString('DBHost');
  cPort          = WideString('DBPort');
  cPassword      = WideString('DBPassword');
  cUser          = WideString('DBUser');
  cDBName        = WideString('DBName');
  cDBODBC        = WideString('ODBC');

begin
  Result := False;

  if ANode = nil then
  begin
    LBLogger.Write(1, 'TSQLDBInfo.LoadConfiguration', lmt_Warning, 'No node!', GetProcessID);
    Exit;
  end;


  _Item := ANode.FindNode(cDBType);
  if _Item <> nil then
    FDBType := TDBType(StrToIntDef(AnsiString(_Item.TextContent), 0));


  case FDBType of
    dbt_Unknown: ;

    dbt_MySQL55:
      begin
        _Item := ANode.FindNode(cHost);
        if _Item <> nil then
          FHost := AnsiString(_Item.TextContent);

        _Item := ANode.FindNode(cPort);
        if _Item <> nil then
          FPort := StrToIntDef(AnsiString(_Item.TextContent), 0);

        _Item := ANode.FindNode(cPassword);
        if _Item <> nil then
          FPassword := AnsiString(_Item.TextContent);

        _Item := ANode.FindNode(cUser);
        if _Item <> nil then
          FUser := AnsiString(_Item.TextContent);

        _Item := ANode.FindNode(cDBName);
        if _Item <> nil then
          FDBName := AnsiString(_Item.TextContent);
      end;

    dbt_SQLite:
      begin
        _Item := ANode.FindNode(cDBName);
        if _Item <> nil then
          FDBName := AnsiString(_Item.TextContent);
      end;

    dbt_ODBC:
      begin
        _Item := ANode.FindNode(cDBODBC);
        if _Item <> nil then
          FODBC := AnsiString(_Item.TextContent);

        _Item := ANode.FindNode(cPassword);
        if _Item <> nil then
          FPassword := AnsiString(_Item.TextContent);

        _Item := ANode.FindNode(cUser);
        if _Item <> nil then
          FUser := AnsiString(_Item.TextContent);
      end;
  end;

  Result := True;
end;

function TSQLDBInfo.getConnection: TSQLConnection;
var
  _ATransaction : TSQLTransaction = nil;
  _NewDB        : Boolean = false;
  _FilePath     : String = '';

begin
  Result := nil;

  if not Self.DataComplete then
  begin
    LBLogger.Write(1, 'TSQLDBInfo.getConnection', lmt_Warning, 'Cannot create connection: uncompleted data!', GetProcessID);
    Exit;
  end;

  try
    case FDBType of

      dbt_MySQL55:
        begin
          Result := TMySQL55Connection.Create(nil);
          Result.HostName := FHost;
          (Result as TMySQL55Connection).Port := FPort;
          Result.DatabaseName := FDBName;
          Result.UserName  := FUser;
          Result.Password := FPassword;
        end;

      dbt_SQLite:
        begin
          _NewDB := not FileExists(FDBName);

          if _NewDB then
          begin
            _FilePath := ExtractFilePath(FDBName);
            if not DirectoryExists(_FilePath) then
              ForceDirectories(_FilePath);
          end;

          Result := TSQLite3Connection.Create(nil);
          Result.DatabaseName := FDBName;
          Result.Password := FPassword;
        end;

      dbt_ODBC:
        begin
          Result := TODBCConnection.Create(nil);
          Result.DatabaseName := FODBC;

          if Length(FUser) > 0 then
            Result.UserName := FUser;

          if Length(FPassWord) > 0 then
            Result.Password := FPassWord;
        end;
    end;

    if Result <> nil then
    begin
      Result.Open();
      if FDBType = dbt_SQLite then
      begin
        if sqlite3_busy_timeout((Result as TSQLite3Connection).Handle, 10000) <> SQLITE_OK then
        begin
          LBLogger.Write(1, 'TSQLConnectionManager.Connect', lmt_Warning, 'SQLite busy!', GetProcessID);
          FreeAndNil(Result);
        end;
      end;
    end;

  except
    on E : Exception do
    begin
      if Result <> nil then
        FreeAndNil(Result);

      LBLogger.Write(1, 'TSQLConnectionManager.Connect', lmt_Error, PChar(Format('Error: <%s>', [E.Message])), GetProcessID);
    end;
  end;

  if Result <> nil then
  begin
    if not Result.Connected then
    begin
      LBLogger.Write(1, 'TSQLConnectionManager.Connect', lmt_Warning, 'Database not connected! Closing connection ...', GetProcessID);
      FreeAndNil(Result);
    end
    else begin
      _ATransaction := TSQLTransaction.Create(nil);
      _ATransaction.DataBase := Result;
      Result.Transaction := _ATransaction;
      if _NewDB then
      begin
        try
          Self.DoNewDatabase(Result);

        except
          on E : Exception do
          begin
            FreeAndNil(Result);
            _ATransaction.Free;
            LBLogger.Write(1, 'TSQLConnectionManager.Connect', lmt_Error, PChar(Format('Error: <%s>', [E.Message])), GetProcessID);
          end;
        end;
      end;
    end;
  end
  else
    LBLogger.Write(1, 'TSQLConnectionManager.Connect', lmt_Warning,
                  PChar(Format('Wrong connection data! ' +
                               'Host: <%s>  -  Port: <%d>  -  User: <%s>  -  DBName: <%s>',
                               [FHost, FPort, FUser, FDBName])), GetProcessID);

end;


end.

