unit ULBLogger;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils;

type
  TLBLoggerMessageType = (lmt_Unknown = 0,
                          lmt_Error   = 1,
                          lmt_Info    = 2,
                          lmt_Debug   = 3,
                          lmt_Warning = 4);

  TLBLoggerMessageTypeSet = set of TLBLoggerMessageType;


  ILBLogger = Interface(IUnknown)
    ['{1019E32D-D1EF-E411-938A-543530D8EA6B}']

    function Write(LogLevel: Byte; Sender: PChar; MsgType: TLBLoggerMessageType; MsgText: PChar; ThreadId: Int64): Boolean; stdcall;

    procedure set_Filename(AFilename: PChar); stdcall;
    procedure set_Folder(AFolder: PChar); stdcall;
    procedure set_MaxFileSize(ASize: Int64); stdcall;
    procedure set_MaxLogLevel(AValue: Byte); stdcall;
    procedure set_EnabledMessageTypes(AList: PChar); stdcall;  // The string '1,3,4' will enable error, debug and warning messages
  end;
  
 { TLBLogger }

 TLBLogger = class(TInterfacedObject, ILBLogger)
   strict private
     FMaxLogLevel : Byte;
     FFileName : AnsiString;
     FFilePath : AnsiString;
     FMessageList : TThreadList;
     FWriter : TObject;
     FMaxFileSize : Int64;
     FEnabledMessages : TLBLoggerMessageTypeSet;
     procedure CreateWriter;
     procedure setFileName(AValue : AnsiString);
     procedure setFilePath(AValue : AnsiString);
     procedure setMaxLogLevel(AValue: Byte);
   public
     constructor Create;
     destructor Destroy; override;

     // Interface
     function Write(LogLevel: Byte; Sender: PChar; MsgType: TLBLoggerMessageType; MsgText: PChar; ThreadId: Int64 = 0): Boolean; stdcall;
     procedure set_Filename(AFilename: PChar); stdcall;
     procedure set_Folder(AFolder: PChar); stdcall;
     procedure set_MaxFileSize(ASize: Int64); stdcall;
     procedure set_MaxLogLevel(AValue: Byte); stdcall;
     procedure set_EnabledMessageTypes(AList: PChar); stdcall;
     // ------------------------------------------------------------------------------------------------

     property FileName: AnsiString read FFileName write setFileName;
     property FilePath: AnsiString read FFilePath write setFilePath;
     property MaxFileSize: Int64 read FMaxFileSize write set_MaxFileSize;
     property MaxLogLevel: Byte read FMaxLogLevel write setMaxLogLevel;
     property EnabledMessages: TLBLoggerMessageTypeSet write FEnabledMessages;
 end;


var
  LBLogger : TLBLogger = nil;

implementation

const
  cLBLoggerMessagePrefix : array [TLBLoggerMessageType] of String = ('#UNK# - ',
                                                                     '#ERR# - ',
                                                                     '#INF# - ',
                                                                     '#DBG# - ',
                                                                     '#WRN# - ');


type

  { TLBLoggerWriter }
  pLBLoggerWriter = ^TLBLoggerWriter;
  TLBLoggerWriter = class(TThread)
    strict private
      FReference : pLBLoggerWriter;
      FMessageList : TThreadList;
      FFileName : String;
      FFilePath : String;
      FDateTimeFormat : String;
      FMaxFileSize : Int64;
      procedure ChangeLogFileName();
      procedure CreateLockFile();
    protected
      procedure Execute; override;
    public
      constructor Create;
      destructor Destroy; override;

      property Reference: pLBLoggerWriter write FReference;
      property MessageList: TThreadList write FMessageList;
      property FileName: String write FFileName;
      property FilePath: String write FFilePath;
      property MaxFileSize: Int64 write FMaxFileSize;
  end;


  { TLBLoggerMessage }

  TLBLoggerMessage = class(TObject)
    strict private
      FMessage : AnsiString;
      FCallingRoutine : AnsiString;
      FTime : TDateTime;
      FMsgType : TLBLoggerMessageType;
      FThreadId : Int64;
      procedure Clear;
    public
      constructor Create;
      property Message : AnsiString read FMessage write FMessage;
      property CallingRoutine : AnsiString read FCallingRoutine write FCallingRoutine;
      property Time : TDateTime read FTime write FTime;
      property MsgType : TLBLoggerMessageType read FMsgType write FMsgType;
      property ThreadId : Int64 read FThreadId write FThreadId;

  end;

const
  cLockFileNameSuffix = String('.lck');


procedure TLBLogger.CreateWriter;
var
  _Writer : TLBLoggerWriter;
begin
  if FWriter = nil then
  begin
    _Writer := TLBLoggerWriter.Create();
    FWriter := _Writer;
    _Writer.Reference := @FWriter;
    _Writer.MessageList := FMessageList;
    _Writer.FileName := FFileName;
    _Writer.FilePath := FFilePath;
    _Writer.MaxFileSize := FMaxFileSize;
    _Writer.Start();
  end;
end;

procedure TLBLogger.setFileName(AValue: AnsiString);
begin
  if Self = nil then
    Exit;

  if FFileName <> AValue then
  begin
    FFileName := AValue;
    if FWriter <> nil then
      (FWriter as TLBLoggerWriter).FileName := FFileName;
  end;
end;

procedure TLBLogger.setFilePath(AValue: AnsiString);
begin
  if Self = nil then
    Exit;

  if FFilePath <> AValue then
  begin
    FFilePath := IncludeTrailingPathDelimiter(AValue);

    if not DirectoryExists(FFilePath) then
      ForceDirectories(FFilePath);

    if FWriter <> nil then
      (FWriter as TLBLoggerWriter).FilePath := FFilePath;
  end;
end;

procedure TLBLogger.setMaxLogLevel(AValue: Byte);
begin
  if Self = nil then
    Exit;

  if FMaxLogLevel <> AValue then
  begin
    Self.Write(1, 'TLBLogger.setMaxLogLevel', lmt_Info, PChar(Format('Changing log level from %d to %d', [FMaxLogLevel, AValue])), GetProcessID);
    FMaxLogLevel := AValue;
  end;
end;

constructor TLBLogger.Create;
begin
  FMessageList := TThreadList.Create;
  FWriter := nil;
  FEnabledMessages := [lmt_Unknown, lmt_Error, lmt_Info, lmt_Debug, lmt_Warning];
  FMaxFileSize := 2 * 1024 * 1024; // 2 MB
  FMaxLogLevel := 3;

  FFileName := '';
  FFilePath := '';
end;

destructor TLBLogger.Destroy;
var
  _List : TList;
  _AMessage : TLBLoggerMessage;

begin
  if FWriter <> nil then
  begin
     (FWriter as TLBLoggerWriter).Terminate;
     while (FWriter <> nil) do
       Sleep(20);
  end;

  if FMessageList <> nil then
  begin
    try
      _List := FMessageList.LockList;
      while _List.Count > 0 do
      begin
        try
          _AMessage := TLBLoggerMessage(_List.Extract(_List.Items[0]));
        except
          _AMessage := nil;
        end;
        if _AMessage <> nil then
          FreeAndNil(_AMessage);
      end;
    finally
      FMessageList.UnlockList;
    end;

    try
      FreeAndNil(FMessageList);
    except
    end;
  end;
  inherited Destroy;
end;

function TLBLogger.Write(LogLevel: Byte; Sender: PChar; MsgType: TLBLoggerMessageType; MsgText: PChar; ThreadId: Int64 = 0): Boolean; stdcall;
var
  _List : TList = nil;
  _AMessage : TLBLoggerMessage = nil;
begin
  if Self = nil then
    Exit;

  Result := False;
  if (LogLevel <= FMaxLogLevel) and (MsgType in FEnabledMessages) then
  begin
    _AMessage := TLBLoggerMessage.Create;
    _AMessage.Time := Now();
    _AMessage.MsgType := MsgType;
    _AMessage.Message := MsgText;
    _AMessage.ThreadId := ThreadId;
    _AMessage.CallingRoutine := Sender;

    try
      _List := FMessageList.LockList;
      _List.Add(Pointer(_AMessage));
    finally
      FMessageList.UnlockList;
      _List := nil;
    end;

    CreateWriter;

    Result := True;
  end;
end;

procedure TLBLogger.set_Filename(AFilename: PChar); stdcall;
begin
  setFileName(strpas(AFilename));
end;

procedure TLBLogger.set_Folder(AFolder: PChar); stdcall;
begin
  setFilePath(strpas(AFolder));
end;

procedure TLBLogger.set_MaxFileSize(ASize: Int64); stdcall;
begin
  if Self = nil then
    Exit;

  if FMaxFileSize <> ASize then
  begin
    FMaxFileSize := ASize;
    if FWriter <> nil then
      (FWriter as TLBLoggerWriter).MaxFileSize := FMaxFileSize;
  end;
end;

procedure TLBLogger.set_MaxLogLevel(AValue: Byte); stdcall;
begin
  FMaxLogLevel := AValue;
end;

procedure TLBLogger.set_EnabledMessageTypes(AList: PChar); stdcall;
var
  _sMessages : String;
  _Messages : TStringList = nil;
  _Value : Byte;
  i : Integer;

begin
  if Self = nil then
    Exit;

  try
    _sMessages := strpas(AList);
    if Length(AList) > 0 then
    begin
      FEnabledMessages := [];
      _Messages := TStringList.Create;
      _Messages.StrictDelimiter := True;
      _Messages.Delimiter := ',';
      _Messages.DelimitedText := _sMessages;
      if _Messages.Count > 0 then
      begin
        for i := 0 to _Messages.Count - 1 do
        begin
          _Value := StrToIntDef(Trim(_Messages.Strings[i]), 0);
          if _Value > 0 then
            FEnabledMessages += [TLBLoggerMessageType(_Value)];
        end;
      end;
    end;

  finally
    if _Messages <> nil then
      _Messages.Free;
  end;
end;

{ TLBLoggerWriter }

procedure TLBLoggerWriter.ChangeLogFileName;
var
  _FileSize : Int64;
  _FBytes : File of Byte;

begin
  if FileExists(FFilePath + PathDelim + FFileName) then
  begin
    _FileSize := 0;
    try
      Assign(_FBytes, FFilePath + PathDelim + FFileName);
      Reset(_FBytes);
      _FileSize := FileSize(_FBytes);
    except
      _FileSize := 0;
    end;
    Close (_FBytes);
    if _FileSize > FMaxFileSize then
      RenameFile(FFilePath + PathDelim + FFileName, FFilePath + PathDelim + FFileName + FormatDateTime('yyyyddmm_hhnnss', Now()));
  end;
end;

procedure TLBLoggerWriter.CreateLockFile;
var
  _Handle : THandle;

begin
  if not FileExists(FFilePath + PathDelim + FFileName + cLockFileNameSuffix) then
  begin
    _Handle := FileCreate(FFilePath + PathDelim + FFileName + cLockFileNameSuffix);
    if _Handle <> THandle(-1) then
      FileClose(_Handle);
  end;
end;

procedure TLBLoggerWriter.Execute;
var
  _AMessage : TLBLoggerMessage = nil;
  _FLock : THandle = THandle(-1);
  _List : TList = nil;
  _sThread : AnsiString;
  _Counter : Integer;
  _sMessageType : AnsiString;
  _txtFile : TFileStream = nil;
  _sMessage : AnsiString;
  _MaxMsgs : Integer = 0;
  _nMsgs : Integer = 0;
  _MsgsWritten : Integer = 0;

const
  _cWaitTime = 10000; // 10 sec
  _cInnerWait = 1;
  _cCicles = _cWaitTime div _cInnerWait;

begin
  if (Length(FFileName) > 0) and
     (Length(FFilePath) > 0) then
  begin
    try

      if not DirectoryExists(FFilePath) then
        ForceDirectories(FFilePath);

      // if doesn't exist create lock file
      CreateLockFile();


      repeat
        _MaxMsgs := 0;

        {$REGION 'Number of messages to write'}
        try
          _List := FMessageList.LockList;
          if _List <> nil then
            _MaxMsgs := _List.Count;
        finally
          FMessageList.UnlockList;
        end;
        {$ENDREGION}

        if _MaxMsgs = 0 then
          Break;

        {$REGION 'Opening lock file'}
        // wait until lock file or timeout
        _Counter := 0;
        repeat
          _FLock := FileOpen(FFilePath + PathDelim + FFileName + cLockFileNameSuffix,
                             fmOpenWrite or fmShareExclusive);
          if _FLock <> THandle(-1) then
            Break;

          Inc(_Counter);
          if _Counter >= _cCicles then
             Break;

          Sleep(_cInnerWait);
        until False;
        {$ENDREGION}

        if _FLock <> THandle(-1) then
        begin
          ChangeLogFileName();

          try
            if not FileExists(FFilePath + PathDelim + FFileName) then
              _txtFile := TFileStream.Create(FFilePath + PathDelim + FFileName, fmOpenReadWrite or fmCreate or fmShareExclusive)
            else
              _txtFile := TFileStream.Create(FFilePath + PathDelim + FFileName, fmOpenReadWrite or fmShareExclusive);

            _txtFile.Seek(0, soFromEnd);

            _MsgsWritten := 0;

            repeat
              {$REGION 'Writing messages'}
              // Vengono scritti i messaggi in coda fino ad un massimo pari al numero
              // di messaggi presenti al momento che precede il lock del file
              _AMessage := nil;
              _nMsgs := 0;

              try
                _List := FMessageList.LockList;
                if (_List <> nil) then
                begin
                  _nMsgs := _List.Count;
                  if _nMsgs > 0 then
                    _AMessage := TLBLoggerMessage(_List.Extract(_List.Items[0]));
                end;
              finally
                FMessageList.UnlockList;
              end;

              if _nMsgs = 0 then
                Break;

              if _AMessage <> nil then
              begin

                if _AMessage.ThreadId > 0 then
                  _sThread := '#' + IntToStr(_AMessage.ThreadId) + '# - '
                else
                  _sThread := '';

                if _AMessage.MsgType IN [lmt_Unknown .. lmt_Warning] then
                  _sMessageType := cLBLoggerMessagePrefix[_AMessage.MsgType]
                else
                  _sMessageType := '#???#';

                _sMessage := FormatDateTime(FDateTimeFormat, _AMessage.Time) + ' - ' +
                             _sMessageType +
                             _sThread +
                             _AMessage.CallingRoutine + ' - ' +
                             _AMessage.Message + sLineBreak;

                _txtFile.WriteBuffer(Pointer(@_sMessage[1])^, Length(_sMessage));

                FreeAndNil(_AMessage);

              end;

              Inc(_MsgsWritten);
              if _MsgsWritten >= _MaxMsgs then
                Break;

              {$ENDREGION}
            until false;

          except
          end;

          if _txtFile <> nil then
             FreeAndNil(_txtFile);

          FileClose(_FLock);
          _FLock := THandle(-1);

          Sleep(5);
        end;
      until false;

    finally
      if _FLock <> THandle(-1) then
        FileClose(_FLock);
    end;
  end;
end;

constructor TLBLoggerWriter.Create();
begin
  inherited Create(True);
  FreeOnTerminate := True;
  FReference := nil;
  FMaxFileSize := 2 * 1024 * 1024;
  FDateTimeFormat := 'dd/mm/yy hh.nn.ss.zzz';
end;

destructor TLBLoggerWriter.Destroy;
begin
  if FReference <> nil then
    FReference^ := nil;

  inherited Destroy;
end;

{ TLBLoggerMessage }

procedure TLBLoggerMessage.Clear;
begin
  FMessage := '';
  FCallingRoutine := '';
  FTime := 0;
  FThreadId := 0;
  FMsgType := lmt_Unknown;
end;

constructor TLBLoggerMessage.Create;
begin
  Clear;
end;


end.

