unit uLBUtils;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils;

type

  { TObjectSerializeDeserialize }

  TObjectSerializeDeserialize = class helper for TObject
    function FillFromJSON(AJSONString: String): Boolean;
    function SerializeAsJSON(): UTF8String;
  end;

  { TCString }

  TCString = class(TObject)
    strict private
      FString : PChar;
      function get_AsString: String;
      function get_Length: Integer;
      procedure set_AsString(AValue: String);
    public
      constructor Create;
      destructor Destroy; override;

      property AsString: String read get_AsString write set_AsString;
      property AsCString: pChar read FString;
      property Len: Integer read get_Length;
  end;

  function getTemporaryFolder(): String;
  function FindFilesInFolder(Path: String; Mask: String; Recursive: Boolean; Files: TStringList): Boolean;

  function generateGUID(): String;

  function SpecialCharToHTML(AString: String): String;

implementation

uses
  fpjsonrtti, fpjson, strutils, ULBLogger;


function generateGUID(): String;
var
  _GUID : TGuid;
begin
  Result := '';
  if CreateGUID(_GUID) = 0 then
    Result := GUIDToString(_GUID);
end;


function SpecialCharToHTML(AString: String): String;
begin
  Result := '';

  if Length(AString) > 0 then
  begin
    Result := AnsiReplaceStr(AString, #176, '&deg;');
    Result := AnsiReplaceStr(Result, 'à', '&agrave;');
    Result := AnsiReplaceStr(Result, 'ò', '&ograve;');
    Result := AnsiReplaceStr(Result, 'è', '&egrave;');
    Result := AnsiReplaceStr(Result, 'ù', '&ugrave;');
    Result := AnsiReplaceStr(Result, 'ì', '&igrave;');
  end;
end;

function FindFilesInFolder(Path: String; Mask: String; Recursive: Boolean; Files: TStringList): Boolean;
var
  _Search : TSearchRec;

begin
  Result := false;

  if Files = nil then
    Exit;

  if DirectoryExists(Path) then
  begin
    Path := IncludeTrailingPathDelimiter(Path);
    try
      if FindFirst(Path + Mask, faAnyFile, _Search) = 0 then
      begin
        repeat
          if (_Search.Attr and faDirectory > 0) then
          begin
            if Recursive and (_Search.Name <> '.') and (_Search.Name <> '..') then
              Result := FindFilesInFolder(Path + _Search.Name, Mask, Recursive, Files);
				  end
          else begin
            Files.Add(Path + _Search.Name);
            Result := True;
					end;
				until FindNext(_Search) <> 0;
			end;
		finally
      SysUtils.FindClose(_Search);
		end;
	end;
end;


function getTemporaryFolder: String;
begin
  Result := GetTempDir(False);

  if Length(Result) = 0 then
    Result := GetTempDir(True);

  if Length(Result) > 0 then
    Result := IncludeTrailingPathDelimiter(Result);
end;

{ TCString }

function TCString.get_AsString: String;
begin
  Result := strpas(FString);
end;

function TCString.get_Length: Integer;
begin
  Result := strlen(FString);
end;

procedure TCString.set_AsString(AValue: String);
begin
  if Self = nil then
    Exit;

  try

    if FString <> nil then
    begin
      StrDispose(FString);
      FString := nil;
    end;

    FString := StrAlloc(Length(AValue)  + 1);
    if Length(AValue) > 0 then
      strpcopy(FString, AValue)
    else
      FillChar(FString^, 1, 0);

  except
  end;
end;

constructor TCString.Create;
begin
  inherited Create;
  FString := nil;
end;

destructor TCString.Destroy;
begin
  if FString <> nil then
    StrDispose(FString);

  inherited Destroy;
end;


{ TObjectSerializeDeserialize }

function TObjectSerializeDeserialize.FillFromJSON(AJSONString: String): Boolean;
var
  _JSonDeStreamer : TJSONDeStreamer = nil;
begin
  Result := False;

  if Self = nil then
  begin
    LBLogger.Write(1, 'TObjectSerializeDeserialize.FillFromJSON', lmt_Warning, 'Null object!', GetProcessID);
    Exit;
  end;

  if Length(AJSONString) = 0 then
  begin
    LBLogger.Write(1, 'TObjectSerializeDeserialize.FillFromJSON', lmt_Warning, 'Invalid JSON string!', GetProcessID);
    Exit;
  end;

  try
    _JSonDeStreamer := TJSONDeStreamer.Create(nil);
    _JSonDeStreamer.JSONToObject(AJsonString, Self);
    Result := True;
  except
    on E: Exception do
      LBLogger.Write(1, 'TObjectSerializeDeserialize.FillFromJSON', lmt_Error, PChar(E.Message), GetProcessID);
  end;

  if _JSonDeStreamer <> nil then
    _JSonDeStreamer.Free;
end;

function TObjectSerializeDeserialize.SerializeAsJSON(): UTF8String;
var
  _JsonStreamer : fpjsonrtti.TJSONStreamer = nil;
  _JsonData     : fpjson.TJSONData = nil;
begin
  Result := '';

  if Self = nil then
    Exit;

  try
    _JsonStreamer         := fpjsonrtti.TJSONStreamer.Create(nil);
    _JsonStreamer.Options := _JsonStreamer.Options + [jsoEnumeratedAsInteger, jsoSetEnumeratedAsInteger];
    _JsonData             := _JsonStreamer.ObjectToJSON(Self);
    Result                := UTF8Encode(_JsonData.AsJSON);
  except
    on e: Exception do
    begin
      LBLogger.Write(1, 'TObjectSerializeDeserialize.SerializeAsJSON', lmt_Error, PChar(E.Message), GetProcessID);
      Result := '';
    end;
  end;

  if _JsonData <> nil then
    _JsonData.Free;

  if _JsonStreamer <> nil then
    _JsonStreamer.Free;
end;


end.

